package uulm.teamname.marvelous.gamelibrary.config;

import com.fasterxml.jackson.annotation.JsonIgnore;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotEmpty;
import jakarta.validation.constraints.Size;
import uulm.teamname.marvelous.gamelibrary.Tuple;

import java.util.*;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;

/**
 * POJO describing the CharacterConfig as defined by the standard document.
 */
public class CharacterConfig {

    @NotEmpty
    @Valid
    @Size(min = 24, message = "has less than 24 characters")
    public CharacterProperties[] characters;

    @JsonIgnore private Map<String, CharacterProperties> propertyMap;
    @JsonIgnore private Map<String, CharacterProperties> unmodifiablePropertyMap;
    @JsonIgnore private Map<Integer, CharacterProperties> idToCharacter;

    /**
     * @return a unmodifiable {@link Map}<{@link String}, {@link CharacterProperties}> containing all properties.
     * If not yet existent, initialize the Map
     */
    @JsonIgnore
    public Map<String, CharacterProperties> getMap() {
        // lazy initialization
        if (propertyMap == null) {
            propertyMap = new HashMap<>();
            for (CharacterProperties property: characters) {
                propertyMap.put(property.name, property);
            }
            unmodifiablePropertyMap = Collections.unmodifiableMap(propertyMap);
        }

        return unmodifiablePropertyMap;
    }

    /**
     * @return a {@link Map}<{@link Integer}, {@link CharacterProperties}> from characterID to character.
     * If not yet existent, initialize the Map
     */
    @JsonIgnore
    public Map<Integer, CharacterProperties> getIDMap() {
        // lazy initialization
        if (idToCharacter == null) {
            idToCharacter = new HashMap<>();
            for (CharacterProperties property: characters) {
                idToCharacter.put(property.characterID, property);
            }
        }

        return idToCharacter;
    }

    @JsonIgnore
    public Tuple<CharacterProperties[], CharacterProperties[]> getDisjointSetsOfPropertiesOfSize(int size) {
        if ((size * 2) > characters.length) {
            return null;
        } else {
            var characterList = Arrays.asList(this.characters);
            Collections.shuffle(characterList);
            return Tuple.of(characterList.subList(0, size).toArray(new CharacterProperties[0]),
                    characterList.subList(size, 2 * size).toArray(new CharacterProperties[0]));
        }
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        CharacterConfig that = (CharacterConfig) o;
        return Arrays.equals(characters, that.characters) && Objects.equals(propertyMap, that.propertyMap) && Objects.equals(unmodifiablePropertyMap, that.unmodifiablePropertyMap);
    }

    @Override
    public int hashCode() {
        int result = Objects.hash(propertyMap, unmodifiablePropertyMap);
        result = 31 * result + Arrays.hashCode(characters);
        return result;
    }
}
