package uulm.teamname.marvelous.gamelibrary.events;

import uulm.teamname.marvelous.gamelibrary.IntVector2;
import uulm.teamname.marvelous.gamelibrary.entities.Entity;
import uulm.teamname.marvelous.gamelibrary.entities.EntityID;
import uulm.teamname.marvelous.gamelibrary.entities.StoneType;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.StringJoiner;

/**
 * A class made for building all sorts of {@link Event Events} as nicely as possible.
 * It works by taking .with[KEY]() functions and then building a specific event at the end.
 */
public class EventBuilder {
    private EventType type;

    // Keys used primarily in EntityEvents, but also in others
    private EntityID targetEntity;
    private IntVector2 targetField;
    private Integer amount;
    private Entity entity;

    // Keys used primarily in CharacterEvents
    private EntityID originEntity;
    private IntVector2 originField;
    private StoneType stoneType;

    // Keys used primarily in GameEvents
    private Integer roundCount = null;
    private Integer turnCount = null;
    private EntityID[] characterOrder;
    private EntityID nextCharacter;

    private Integer playerWon;

    private String message;
    private Integer timeLeft;

    // Keys used primarily in GameStateEvents
    private Entity[] entities;
    private EntityID[] turnOrder;
    private IntVector2 mapSize;
    private EntityID activeCharacter;
    private Integer[] stoneCooldowns;
    private Boolean winCondition;

    // Keys used in TeleportedEvents (plus origin/targetFields)
    public EntityID teleportedEntity;
    public EntityID originPortal, targetPortal;

    // Keys used in CustomEvents
    private String teamIdentifier;
    private HashMap<String, Object> customContent;

    /**
     * Creates a new {@link EventBuilder} used for building {@link Event Events}.
     * @param eventType is the type of Event that the final event will have
     */
    public EventBuilder(EventType eventType) {
        this.type = eventType;
    }

    /**
     * @deprecated Use {@link #EventBuilder(EventType)} instead.
     */
    @Deprecated
    public EventBuilder withType(EventType type) {
        this.type = type;
        return this;
    }

    public EventBuilder withTargetEntity(EntityID targetEntity) {
        this.targetEntity = targetEntity;
        return this;
    }

    public EventBuilder withTargetField(IntVector2 targetField) {
        this.targetField = targetField;
        return this;
    }

    public EventBuilder withAmount(Integer amount) {
        this.amount = amount;
        return this;
    }

    public EventBuilder withEntity(Entity entity) {
        this.entity = entity;
        return this;
    }

    public EventBuilder withOriginEntity(EntityID originEntity) {
        this.originEntity = originEntity;
        return this;
    }

    public EventBuilder withOriginField(IntVector2 originField) {
        this.originField = originField;
        return this;
    }

    public EventBuilder withStoneType(StoneType stoneType) {
        this.stoneType = stoneType;
        return this;
    }

    public EventBuilder withRoundCount(Integer roundCount) {
        this.roundCount = roundCount;
        return this;
    }

    public EventBuilder withTurnCount(Integer turnCount) {
        this.turnCount = turnCount;
        return this;
    }

    public EventBuilder withCharacterOrder(EntityID[] characterOrder) {
        this.characterOrder = characterOrder;
        return this;
    }

    public EventBuilder withNextCharacter(EntityID nextCharacter) {
        this.nextCharacter = nextCharacter;
        return this;
    }

    public EventBuilder withPlayerWon(Integer playerWon) {
        this.playerWon = playerWon;
        return this;
    }

    public EventBuilder withMessage(String message) {
        this.message = message;
        return this;
    }

    public EventBuilder withTimeLeft(Integer timeLeft) {
        this.timeLeft = timeLeft;
        return this;
    }

    public EventBuilder withEntities(Entity[] entities) {
        this.entities = entities;
        return this;
    }

    public EventBuilder withTurnOrder(EntityID[] turnOrder) {
        this.turnOrder = turnOrder;
        return this;
    }

    public EventBuilder withMapSize(IntVector2 mapSize) {
        this.mapSize = mapSize;
        return this;
    }

    public EventBuilder withActiveCharacter(EntityID activeCharacter) {
        this.activeCharacter = activeCharacter;
        return this;
    }

    public EventBuilder withStoneCooldowns(Integer[] stoneCooldowns) {
        this.stoneCooldowns = stoneCooldowns;
        return this;
    }

    public EventBuilder withWinCondition(Boolean winCondition) {
        this.winCondition = winCondition;
        return this;
    }

    public EventBuilder withTeleportedEntity(EntityID teleportedEntity) {
        this.teleportedEntity = teleportedEntity;
        return this;
    }

    public EventBuilder withOriginPortal(EntityID originPortal) {
        this.originPortal = originPortal;
        return this;
    }

    public EventBuilder withTargetPortal(EntityID targetPortal) {
        this.targetPortal = targetPortal;
        return this;
    }

    public EventBuilder withTeamIdentifier(String teamIdentifier) {
        this.teamIdentifier = teamIdentifier;
        return this;
    }

    public EventBuilder withCustomContent(HashMap<String, Object> customContent) {
        this.customContent = customContent;
        return this;
    }

    /**
     * Builds a {@link GameEvent} from the values given to the builder.
     * <ul>
     *  <li>type is the {@link EventType} of the {@link GameEvent}.</li>
     *  <li>roundCount describes the count of rounds in the RoundSetupEvent</li>
     *  <li>turnCount describes the amount of turns taken in Events</li>
     *  <li>characterOrder describes the order in which characters have their turn</li>
     *  <li>nextCharacter describes the next character in the turn order in the TurnEvent</li>
     *  <li>playerWon describes the player that has won the game in the WinEvent</li>
     *  <li>message describes a generic message delivered as a String in several GameEvents. It is optional.</li>
     *  <li>timeLeft describes the time left for a client to act before getting kicked in the TimeoutWarning event</li>
     * </ul>
     *
     * @return a {@link GameEvent} based on the builder
     */
    public GameEvent buildGameEvent() throws IllegalStateException {
        var gameEvent = new GameEvent();
        gameEvent.type = this.type;
        gameEvent.roundCount = this.roundCount;
        gameEvent.turnCount = this.turnCount;
        gameEvent.characterOrder = this.characterOrder;
        gameEvent.nextCharacter = this.nextCharacter;

        gameEvent.playerWon = this.playerWon;

        gameEvent.message = this.message;
        gameEvent.timeLeft = this.timeLeft;
        return gameEvent;
    }

    /**
     * Builds an {@link EntityEvent} from the values given to the builder.
     * <ul>
     *  <li>type is the {@link EventType} of the EntityEvent</li>
     *  <li>targetEntity is the target {@link Entity}</li>
     *  <li>targetField is the target field in the form of an {@link IntVector2}</li>
     *  <li>amount is a generic amount, for example damage taken</li>
     * </ul>
     *
     * @return an {@link EntityEvent} based on the builder
     */
    public EntityEvent buildEntityEvent() throws IllegalStateException {
        var entityEvent = new EntityEvent();
        entityEvent.type = this.type;
        entityEvent.targetEntity = this.targetEntity;
        entityEvent.targetField = this.targetField;
        entityEvent.amount = this.amount;
        entityEvent.entity = this.entity;
        return entityEvent;
    }

    /**
     * Builds a {@link CharacterEvent} from the values given to the builder.
     * <ul>
     *  <li>type is the {@link EventType} of the EntityEvent</li>
     *  <li>originEntity is the origin {@link Entity}</li>
     *  <li>targetEntity is the target {@link Entity}</li>
     *  <li>originField is the field that an action originates in, delivered as an {@link IntVector2}</li>
     *  <li>targetField is the target field in the form of an {@link IntVector2}</li>
     * </ul>
     *
     * @return a {@link CharacterEvent} based on the builder
     */
    public CharacterEvent buildCharacterEvent() throws IllegalStateException {
        var characterEvent = new CharacterEvent();
        characterEvent.type = this.type;
        characterEvent.originEntity = this.originEntity;
        characterEvent.targetEntity = this.targetEntity;
        characterEvent.originField = this.originField;
        characterEvent.targetField = this.targetField;
        characterEvent.amount = this.amount;
        characterEvent.stoneType = this.stoneType;
        return characterEvent;
    }

    /**
     * Builds a generic {@link GamestateEvent} from the values given to the builder.
     * <ul>
     *  <li>type is the {@link EventType} of the {@link GamestateEvent}</li>
     *  <li>entities is an array of {@link Entity Entities}</li>
     *  <li>turnOrder describes the order in which characters take turns</li>
     *  <li>activeCharacter describes the currently active character</li>
     *  <li>describes whether the win condition is in effect</li>
     * </ul>
     *
     * @return a {@link GamestateEvent} based on the builder
     */
    public GamestateEvent buildGameStateEvent() throws IllegalStateException {
        var gameStateEvent = new GamestateEvent();
        gameStateEvent.type = this.type;
        gameStateEvent.entities = this.entities;
        gameStateEvent.turnOrder = this.turnOrder;
        gameStateEvent.mapSize = this.mapSize;
        gameStateEvent.activeCharacter = this.activeCharacter;
        gameStateEvent.stoneCooldowns = this.stoneCooldowns;
        gameStateEvent.winCondition = this.winCondition;
        return gameStateEvent;
    }

    /**
     * Builds a {@link CustomEvent} from the values given to the builder.
     * <ul>
     *  <li>teamIdentifier is a String for identifying a specific custom content</li>
     *  <li>customContent is a {@link HashMap}<{@link String}, {@link Object}> resembling the JSON keys in the event</li>
     * </ul>
     *
     * @return a {@link CustomEvent} based on the builder
     */
    public CustomEvent buildCustomEvent() throws IllegalStateException {
        var customEvent = new CustomEvent();
        customEvent.type = this.type;
        customEvent.teamIdentifier = this.teamIdentifier;
        customEvent.customContent = this.customContent;
        return customEvent;
    }

    /**
     * Builds a {@link TeleportedEvent}, used for describing teleporting Players into Portals.
     * @return the built {@link TeleportedEvent}
     */
    public TeleportedEvent buildTeleportedEvent() {
        var teleportedEvent = new TeleportedEvent();
        teleportedEvent.type = this.type;
        teleportedEvent.originField = this.originField;
        teleportedEvent.targetField = this.targetField;
        teleportedEvent.teleportedEntity = this.teleportedEntity;
        teleportedEvent.originPortal = this.originPortal;
        teleportedEvent.targetPortal = this.targetPortal;
        return teleportedEvent;
    }

    @Override
    public String toString() {
        StringJoiner joiner = new StringJoiner("\n");
        joiner.add("EventBuilder(all) {");
        for(Field field: this.getClass().getDeclaredFields()) {
            try {
                joiner.add(field.getName() + " = " + field.get(this));
            }catch(IllegalAccessException ignore) { }
        }
        joiner.add("}");
        return joiner.toString();
    }

    public String notNullToString() {
        StringJoiner joiner = new StringJoiner("\n");
        joiner.add("EventBuilder(non-null) {");
        for(Field field: this.getClass().getDeclaredFields()) {
            try {
                Object value = field.get(this);
                if(value != null) {
                    joiner.add(field.getName() + " = " + value);
                }
            }catch(IllegalAccessException ignore) { }
        }
        joiner.add("}");
        return joiner.toString();
    }
}
