package uulm.teamname.marvelous.gamelibrary.events;

import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import uulm.teamname.marvelous.gamelibrary.entities.EntityID;
import uulm.teamname.marvelous.gamelibrary.json.ingame.deserialize.EventDeserializer;
import uulm.teamname.marvelous.gamelibrary.json.ingame.serialize.EventSerializer;

import java.util.Arrays;
import java.util.Objects;

/** Represents a game event for: {@link EventType#RoundSetupEvent}, {@link EventType#TurnEvent}, {@link EventType#WinEvent}, {@link EventType#TurnTimeoutEvent}, {@link EventType#TimeoutWarningEvent}, {@link EventType#TimeoutEvent}, {@link EventType#PauseStartEvent}, {@link EventType#PauseStopEvent}, {@link EventType#DisconnectEvent}. */
@JsonDeserialize(using = EventDeserializer.class)
@JsonSerialize(using = EventSerializer.class)
public class GameEvent extends Event {
    public Integer roundCount;
    public Integer turnCount;
    public EntityID[] characterOrder;
    public EntityID nextCharacter;

    public Integer playerWon;

    public String message;
    public Integer timeLeft;

    @Override
    public boolean check() {
        if(!super.check()) {
            return false;
        }

        switch(type) {
            // all of those events do not need any extra values except the EventType
            case PauseStartEvent:
            case PauseStopEvent:
            case TurnTimeoutEvent:
            case DisconnectEvent:
                break;

            // RoundSetupEvents take a RoundCount and a CharacterOrder
            case RoundSetupEvent:
                if (this.roundCount == null || this.characterOrder == null) {
                    return false;
                }
                break;

            // TurnEvents take a TurnCount and a NextCharacter
            case TurnEvent:
                if (this.turnCount == null || this.nextCharacter == null) {
                    return false;
                }
                break;

            // A WinEvent needs to know what player has won
            case WinEvent:
                if (this.playerWon == null) {
                    return false;
                }
                break;

            // TimeoutEvents give a message back. As the only events to do so, this might be removed later on.
            case TimeoutEvent:
                if (this.message == null) {
                    return false;
                }
                break;

            // TimeoutWarnings carry a message and the amount of time left in seconds. The message might disappear.
            case TimeoutWarningEvent:
                if (this.message == null || this.timeLeft == null) {
                    return false;
                }
                break;
        }

        return true;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        if (!super.equals(o)) return false;
        GameEvent gameEvent = (GameEvent) o;
        return Objects.equals(roundCount, gameEvent.roundCount) && Objects.equals(turnCount, gameEvent.turnCount) && Objects.equals(playerWon, gameEvent.playerWon) && Objects.equals(timeLeft, gameEvent.timeLeft) && Arrays.equals(characterOrder, gameEvent.characterOrder) && Objects.equals(nextCharacter, gameEvent.nextCharacter) && Objects.equals(message, gameEvent.message);
    }

    @Override
    public int hashCode() {
        int result = Objects.hash(super.hashCode(), roundCount, turnCount, nextCharacter, playerWon, message, timeLeft);
        result = 31 * result + Arrays.hashCode(characterOrder);
        return result;
    }

    @Override
    public String toString() {
        return "GameEvent{" +
                "eventType=" + this.type +
                ", roundCount=" + roundCount +
                ", turnCount=" + turnCount +
                ", characterOrder=" + Arrays.toString(characterOrder) +
                ", nextCharacter=" + nextCharacter +
                ", playerWon=" + playerWon +
                ", message='" + message + '\'' +
                ", timeLeft=" + timeLeft +
                '}';
    }
}
