package uulm.teamname.marvelous.gamelibrary.json.ingame.deserialize;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.InjectableValues;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import static org.assertj.core.api.Assertions.*;

import uulm.teamname.marvelous.gamelibrary.IntVector2;
import uulm.teamname.marvelous.gamelibrary.entities.*;
import uulm.teamname.marvelous.gamelibrary.entities.Character;
import uulm.teamname.marvelous.gamelibrary.events.*;
import uulm.teamname.marvelous.gamelibrary.config.CharacterConfig;
import uulm.teamname.marvelous.gamelibrary.config.CharacterProperties;

import java.util.*;
import java.util.concurrent.ThreadLocalRandom;

import static org.mockito.Mockito.*;

class EventDeserializerTest {

    ObjectMapper mapper;
    Map<String, CharacterProperties> propertiesMap;

    @BeforeEach
    void beforeEach() {
        mapper = new ObjectMapper();
        propertiesMap = new HashMap<>();

        var config = mock(CharacterConfig.class);
        when(config.getMap()).thenReturn(propertiesMap);

        mapper.setInjectableValues(new InjectableValues
                .Std()
                .addValue("CharacterConfig", config));
    }

    @Test
    void NotificationEventTest() throws JsonProcessingException {
        var ack = new GamestateEvent();
        ack.type = EventType.Ack;

        var jsonRepresentingAck = "{\"eventType\": \"Ack\"}";

        assertThat((GamestateEvent) mapper.readValue(jsonRepresentingAck, Event.class))
                .withFailMessage("Ack was not properly deserialized")
                .isEqualTo(ack);

        var nack = new GamestateEvent();
        nack.type = EventType.Nack;

        var jsonRepresentingNack = "{\"eventType\": \"Nack\"}";

        assertThat((GamestateEvent) mapper.readValue(jsonRepresentingNack, Event.class))
                .withFailMessage("Nack was not properly deserialized")
                .isEqualTo(nack);
    }

    @Test
    void GamestateEventTest() throws JsonProcessingException {

        var alphaConfig = new CharacterProperties();
        alphaConfig.attackRange = 45;
        alphaConfig.rangedDamage = 65;
        alphaConfig.meleeDamage = 12;
        propertiesMap.put("Alpha", alphaConfig);

        var characterAlpha = new Character(
                new EntityID(EntityType.P1, 4),
                new IntVector2(5, 3),
                "Alpha",
                210,
                9,
                3,
                45, // still -1 because no gameConfig object available
                65,
                12);

        characterAlpha.inventory.addStone(StoneType.valueOf(1));
        characterAlpha.inventory.addStone(StoneType.valueOf(4));

        characterAlpha.setPosition(new IntVector2(4, 2));

        var steelWallConfig = new CharacterProperties();
        steelWallConfig.attackRange = 23;
        steelWallConfig.rangedDamage = 34;
        steelWallConfig.meleeDamage = 45;
        propertiesMap.put("Steel Wall", steelWallConfig);

        var characterSteelWall = new Character(
                new EntityID(EntityType.P1, 4),
                new IntVector2(255, 0),
                "Steel Wall",
                123,
                2,
                21,
                23,
                34,
                45);

        characterSteelWall.setPosition(new IntVector2(6, 7));


        var infinityStone = new InfinityStone(
                new EntityID(EntityType.InfinityStones, 3),
                new IntVector2(52, 13),
                StoneType.valueOf(3));

        infinityStone.setPosition(new IntVector2(7, 22));


        var rock = new Rock(
                new EntityID(EntityType.Rocks, 15),
                new IntVector2(51, 31),
                195);

        rock.setPosition(new IntVector2(4, 285));

        var gamestate = new GamestateEvent();
        gamestate.type = EventType.GamestateEvent;
        gamestate.entities = new Entity[] {
                characterAlpha,
                characterSteelWall,
                infinityStone,
                rock
        };
        gamestate.mapSize = new IntVector2(15, 300);
        gamestate.activeCharacter = new EntityID(EntityType.P2, 3);
        gamestate.stoneCooldowns = new Integer[] {1, 1, 3, 1, 4, 2};
        gamestate.turnOrder = new EntityID[] {
                new EntityID(EntityType.P1, 5),
                new EntityID(EntityType.P1, 4),
                new EntityID(EntityType.P1, 3),
                new EntityID(EntityType.P1, 2),
                new EntityID(EntityType.P1, 1),
                new EntityID(EntityType.P1, 0),
                new EntityID(EntityType.P2, 0),
                new EntityID(EntityType.P2, 1),
                new EntityID(EntityType.P2, 2),
                new EntityID(EntityType.P2, 3),
                new EntityID(EntityType.P2, 4),
                new EntityID(EntityType.P2, 5),
        };
        gamestate.winCondition = true;



        var jsonRepresentingGamestate = """
        {
            "eventType": "GamestateEvent",
            "entities": [
                {
                    "entityType": "Character",
                    "name": "Alpha",
                    "PID": 1,
                    "ID": 4,
                    "HP": 210,
                    "MP": 9,
                    "AP": 3,
                    "stones": [1, 4],
                    "position": [4, 2]
                },
                {
                    "entityType": "Character",
                    "name": "Steel Wall",
                    "PID": 1,
                    "ID": 4,
                    "HP": 123,
                    "MP": 2,
                    "AP": 21,
                    "stones": [],
                    "position": [6, 7]
                },
                {
                    "entityType": "InfinityStone",
                    "ID": 3,
                    "position": [7, 22]
                },
                {
                    "entityType": "Rock",
                    "ID": 15,
                    "HP": 195,
                    "position": [4, 285]
                }
            ],
            "turnOrder": [
            {"entityID": "P1", "ID": 5},
            {"entityID": "P1", "ID": 4},
            {"entityID": "P1", "ID": 3},
            {"entityID": "P1", "ID": 2},
            {"entityID": "P1", "ID": 1},
            {"entityID": "P1", "ID": 0},
            {"entityID": "P2", "ID": 0},
            {"entityID": "P2", "ID": 1},
            {"entityID": "P2", "ID": 2},
            {"entityID": "P2", "ID": 3},
            {"entityID": "P2", "ID": 4},
            {"entityID": "P2", "ID": 5}
            ],
            "mapSize": [15, 300],
            "activeCharacter": {"entityID": "P2", "ID": 3},
            "stoneCooldowns": [1, 1, 3, 1, 4, 2],
            "winCondition": true
        }
        """;

        //System.out.println("=== DEBUG OUTPUT ===\n"
        //        + mapper.readValue(jsonRepresentingGamestate, Event.class)
        //        + "\n====================");

        assertThat((GamestateEvent) mapper.readValue(jsonRepresentingGamestate, Event.class))
                .withFailMessage("GamestateEvent was not properly deserialized")
                .isEqualTo(gamestate);
    }

    @Test
    void GameEventsTest() throws JsonProcessingException {
        Iterator<Integer> randomIntegers = ThreadLocalRandom.current().ints().iterator();

        int repetitions = 1000;

        // Keys in a Game Event are:
        // message, roundCount, turnCount, playerWon, timeLeft, characterOrder, nextCharacter

        EventType type;

        final StringBuilder message = new StringBuilder();
        final int messageLength = 200;
        final int minMessageLength = 5;

        int roundCount;

        int turnCount;

        int playerWon;

        int timeLeft;

        EntityID[] characterOrder;
        final HashSet<EntityID> usedCharacters = new HashSet<>();
        StringBuilder jsonRepresentingCharacterOrder = new StringBuilder();

        EntityID nextCharacter;

        for (int i = 0; i < repetitions; i++) {
            // Get a random type of GameEvent
            switch (Math.abs(randomIntegers.next()) % 8) {
                case 0 -> type = EventType.TurnEvent;
                case 1 -> type = EventType.WinEvent;
                case 2 -> type = EventType.TurnTimeoutEvent;
                case 3 -> type = EventType.TimeoutWarningEvent;
                case 4 -> type = EventType.TimeoutEvent;
                case 5 -> type = EventType.PauseStartEvent;
                case 6 -> type = EventType.PauseStopEvent;
                case 7 -> type = EventType.DisconnectEvent;
                default -> throw new IllegalStateException("Unexpected value: " + Math.abs(randomIntegers.next()) % 8);
            }

            // Get a random message string
            int nameLength = Math.abs(randomIntegers.next() % 12);
            for (int j = 0; j < (nameLength) + 3; j++) {
                // append char between A and Z, and make lowercase 50% of the time
                message.append((char) (
                        (65 + Math.abs(randomIntegers.next() % 26))
                                + 32 * Math.abs(randomIntegers.next() % 2)
                ));
            }

            // Get (realistic-seeming) random values for the int variables
            roundCount = Math.abs(randomIntegers.next()) % 1000;
            turnCount = Math.abs(randomIntegers.next()) % 6000;
            playerWon = Math.abs(randomIntegers.next()) % 2 + 1;
            timeLeft = Math.abs(randomIntegers.next()) % 600 + 1; // zero time left wouldn't make sense

            // Get character order
            characterOrder = new EntityID[12];
            for (int j = 0; j < 12; j++) {
                EntityID entityID;
                do { // generate new ids until id was not added before. Inefficient, but whatever.
                    entityID = new EntityID(
                            Math.abs(randomIntegers.next()) % 2 == 1 ? EntityType.P1 : EntityType.P2,
                            Math.abs(randomIntegers.next()) % 6);
                } while (!usedCharacters.add(entityID));
                characterOrder[j] = entityID;
            }

            nextCharacter = characterOrder[Math.abs(randomIntegers.next()) % 12];
            
            var gameEvent = new GameEvent();
            gameEvent.type = type;
            gameEvent.message = message.toString();
            gameEvent.nextCharacter = nextCharacter;
            gameEvent.timeLeft = timeLeft;
            gameEvent.playerWon = playerWon;
            gameEvent.characterOrder = characterOrder;
            gameEvent.turnCount = turnCount;
            gameEvent.roundCount = roundCount;

            // create JSON representation of CharacterOrder

            jsonRepresentingCharacterOrder.append('[').append('\n');
            for (EntityID id: characterOrder) {
                jsonRepresentingCharacterOrder.append('\t')
                        .append(String.format(
                        "{ \"entityID\": \"%s\", \"ID\": %d}"
                        , id.type, id.id))
                        .append(',')
                        .append('\n');
            }
            jsonRepresentingCharacterOrder.setLength(jsonRepresentingCharacterOrder.length() - 2); // remove trailing ,
            jsonRepresentingCharacterOrder.append('\n').append(']');

            var jsonRepresentingGameEvent = """
                    {
                        "eventType": "%s",
                        "message": "%s",
                        "roundCount": %d,
                        "turnCount": %d,
                        "playerWon": %d,
                        "timeLeft": %d,
                        "characterOrder": %s,
                        "nextCharacter": {"entityID": "%s", "ID": %d}
                    }
                    """.formatted(
                    gameEvent.type,
                    gameEvent.message,
                    gameEvent.roundCount,
                    gameEvent.turnCount,
                    gameEvent.playerWon,
                    gameEvent.timeLeft,
                    jsonRepresentingCharacterOrder.toString(),
                    gameEvent.nextCharacter.type, gameEvent.nextCharacter.id
            );

            assertThat((GameEvent) mapper.readValue(jsonRepresentingGameEvent, Event.class))
                    .isEqualTo(gameEvent);

            message.setLength(0);
            jsonRepresentingCharacterOrder.setLength(0);
            usedCharacters.clear();
        }
    }

    @Test
    void teleportedEventDeserializationTest() throws JsonProcessingException {

        var jsonRepresentingTeleportedEvent = """
                {
                "eventType":"TeleportedEvent",
                "teleportedEntity":{"entityID":"P1", "ID":3},
                "originField":[6, 2],
                "targetField":[11, 14],
                "originPortal":{"entityID":"Portals", "ID":0},
                "targetPortal":{"entityID":"Portals", "ID":2}
                }
                """;

        var teleportedEvent = new EventBuilder(EventType.TeleportedEvent)
                .withTeleportedEntity(new EntityID(EntityType.P1, 3))
                .withOriginField(new IntVector2(6, 2))
                .withTargetField(new IntVector2(11, 14))
                .withOriginPortal(new EntityID(EntityType.Portals, 0))
                .withTargetPortal(new EntityID(EntityType.Portals, 2))
                .buildTeleportedEvent();

        assertThat((TeleportedEvent) mapper.readValue(jsonRepresentingTeleportedEvent, Event.class))
                .isEqualTo(teleportedEvent);
    }

}
